// Copyright (C) 2020 Intel Corporation

// SPDX-License-Identifier: MIT

#include <CL/sycl.hpp>
#include <chrono>
using namespace sycl;

extern const int matrixSize = 100;//128;
static const int iterations = 4;



template <typename T>
double run_sycl(
    const std::vector<T>& vecA,
    const std::vector<T>& vecB,
    std::vector<T>& vecC) {

  const int M = matrixSize;
  const int N = matrixSize;
  const int K = matrixSize;

  using ns = std::chrono::nanoseconds;
  ns::rep best_time = std::numeric_limits<ns::rep>::max();

  std::fill(vecC.begin(), vecC.end(), (T)0);

  buffer<T> bufA{vecA};  // M * K elements
  buffer<T> bufB{vecB};  // K * N elements
  buffer<T> bufC{vecC};  // M * N elements

auto handle_async_error = [](exception_list elist) {
for (auto &e : elist) {
try{ std::rethrow_exception(e); }
catch ( sycl::exception& e ) {
std::cout << "ASYNC EXCEPTION!!\n";
std::cout << e.what() << "\n";
}
}
};


  //queue Q;  // Choose any available device
  queue Q{ gpu_selector{}, handle_async_error };

  std::cout << "Running on device: "
            << Q.get_device().get_info<info::device::name>() << "\n";
  std::cout << "MatrixSize:" << matrixSize << "\n";


  for (int i = 0; i < iterations; ++i) {
    auto start = std::chrono::steady_clock::now();
  try{
    Q.submit([&](handler& h) {
      accessor matrixA{bufA, h, read_only};
      accessor matrixB{bufB, h, read_only};
      accessor matrixC{bufC, h, write_only};

// BEGIN CODE SNIP
      h.single_task([=]() {
        for (int m = 0; m < M; m++) {
          for (int n = 0; n < N; n++) {
            T sum = 0;
            for (int k = 0; k < K; k++) {
              sum += matrixA[m * K + k] * matrixB[k * N + n];
            }
            matrixC[m * N + n] = sum;
          }
        }
      });
// END CODE SNIP
    });
  } catch (sycl::exception &e) {
std::cout << "Caught sync SYCL exception: " << e.what() << "\n";
} catch (std::exception &e) {
std::cout << "Caught std exception: " << e.what() << "\n";
} catch (...) {
std::cout << "Caught unknown exception\n";
}

    Q.wait();  // So that we know the kernel has finished before checking time
    auto duration = std::chrono::steady_clock::now() - start;
    auto time = std::chrono::duration_cast<ns>(duration).count();
    std::cout << "time:" << time/1e9 << "\n";
    best_time = std::min(time, best_time);
  }

  double best_seconds = (double)best_time / 1e9;

  return best_seconds;
}

template double run_sycl<float>(
    const std::vector<float>& vecA,
    const std::vector<float>& vecB,
    std::vector<float>& vecC);
